#ifdef WIN32
#include <process.h>
#include "win32.h"
#else
#include <sys/time.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <net/if.h>
#include <pthread.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>
#include <string.h>

#include "common.h"
#include "box.h"
#include "console-server.h"
#include "btkcam-client.h"

bissmap		bisskey[maxbiss];		// BISS & DCW list in biss.cfg
int		bisscnt=0;

char _logm[4096];
char _flogm[5][16384];
unsigned char _log_mask = 0;

pthread_mutex_t _log_display_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t _log_transaction_mutex = PTHREAD_MUTEX_INITIALIZER;

console_server *remote_console_server = NULL;
btkcam_client *btk[10];

// ---------------------------------- global configuration settings ------------------------------------------
char binary_file_path[256] = {0};
char config_file_path[256] = {0};

int config_background_execution = 0;
int config_debug_level = 1;
int config_log_to_console = 0;
int config_log_to_file = 0;
int config_mode = 2;		// 1-dcw   2-ecm  
char config_logfile_name[256] = {0};
int config_console_logs_broadcast = 0;
int config_console_logs_broadcast_port;

// -------------------------------------- Btkcam -------------------
char	Blines[10][255];
int 	Bcnt=0;


void config_read_fix_dos_style( char *config_line )
{
	int length = strlen( config_line );
	if( length > 0 )
	{
		if( config_line[length - 1] == (unsigned char)0x0D || config_line[length - 1] == (unsigned char)0x0A)
			config_line[length - 1] = 0;
		
		if( length > 1 )
		{
			if( config_line[length - 2] == (unsigned char)0x0D || config_line[length - 2] == (unsigned char)0x0A)
				config_line[length - 2] = 0;
		}
	}
}

int config_read_settings( void )
{
	FILE *config_file;
	char config_line[1024] = {0};
	char url_line[300] = {0};
	char urluser[20];
	char urlpass[20];
	int items_read,ic;
	unsigned int temp_int;
	
	
	//get the binary file path
	readlink("/proc/self/exe", binary_file_path, sizeof(binary_file_path));
	ic=strlen(binary_file_path)-1;
	while (binary_file_path[ic] != '/') ic--;
	binary_file_path[ic+1]=0;

	strcpy(config_file_path,binary_file_path);
	strcat(config_file_path,"btkcam.cfg");
	if (!fexists(config_file_path)) strcpy(config_file_path,"/var/keys/btkcam.cfg");
		

	config_file = fopen(config_file_path, "rt");
	if (config_file == NULL) 
		{
		_slog(NO_TRANS, _DEBUG_BASIC, 0, "ERROR: Copy Btkcam.cfg into /var/keys or binary folder.", 2, true);
		return 1;
		}
	
	
	while( fgets(config_line, 1024, config_file) )
	{
	config_read_fix_dos_style( config_line );
	if( !strncmp("Box: ", config_line, 5) )  sscanf(config_line,"Box: %d",&box::type);
	if( !strncmp("Back: ", config_line, 6) )  sscanf(config_line,"Back: %d",&config_background_execution);
	if( !strncmp("Mode: ", config_line, 6) )  sscanf(config_line,"Mode: %d",&config_mode);
	if( !strncmp("Dbg_Lvl: ", config_line, 9) ) sscanf(config_line,"Dbg_Lvl: %d",&config_debug_level);
	if( !strncmp("Log_Con: ", config_line, 9) ) sscanf(config_line,"Log_Con: %d",&config_log_to_console); 
	if( !strncmp("Log_File: ", config_line, 10) )  sscanf(config_line,"Log_File: %s %d",config_logfile_name, &config_log_to_file); 
	if( !strncmp("Log_Tcp: ", config_line, 9)) sscanf(config_line,"Log_Tcp: %d %d",&config_console_logs_broadcast_port, &config_console_logs_broadcast); 
	if( !strncmp("B: ", config_line, 3))  	if (Bcnt<11) memcpy(Blines[Bcnt++],config_line,strlen(config_line)); 
						else printf("Too many B-lines. Only first 10 will be active.\n");
	}
	fclose( config_file );

	//read biss
	strcpy(config_file_path,"/var/keys/biss.cfg");
	config_file = fopen(config_file_path, "rt");
	if (config_file != NULL) 
		{
		while( fgets(config_line, 1024, config_file) )
			if (strncmp(config_line,"B2: ",4)==0 && bisscnt<maxbiss) 
			{
			sscanf(config_line, "B2: %x %hx %hx %hx %hx %s",&bisskey[bisscnt].ns, &bisskey[bisscnt].pmt, &bisskey[bisscnt].sid,
					&bisskey[bisscnt].vpid,&bisskey[bisscnt].apid,bisskey[bisscnt].keys);
			//printf("Biss (%03d) NS:%08X PMT:%04X SID:%04X PIDS: %04X-%04X : %s\n",bisscnt,bisskey[bisscnt].ns,bisskey[bisscnt].pmt,bisskey[bisscnt].sid,bisskey[bisscnt].vpid,bisskey[bisscnt].apid,bisskey[bisscnt].keys);
			bisscnt++;
			}
		fclose( config_file );
		}
	
	return 0;
}



void config_show_settings( char *settings )
{
	char buffer[1024];
	char line[128];
	int  ixx;
	sprintf(line, "\nBtkcam Configuration settings:\n"); strcpy( buffer, line );
	sprintf(line, "  Box             : %s\n", box::type == 0 ? "Dreambox" : box::type == 1 ? "Triple Dragon" : box::type == 2 ? "DGStation Device" : box::type == 3 ? "Dbox2" : "Invalid"); strcat( buffer, line );
	sprintf(line, "  Background EXE  : %s\n", config_background_execution ? "Yes" : "No"); strcat( buffer, line );
	sprintf(line, "  Debug level     : %i\n", config_debug_level); strcat( buffer, line );
	sprintf(line, "  Log to Console  : %s\n", config_log_to_console ? "Yes" : "No"); strcat( buffer, line );
	sprintf(line, "  Log to file     : %s %s\n", config_log_to_file ? "Yes" : "No",config_logfile_name ); strcat( buffer, line );
	sprintf(line, "  Log to Network  : %s %d\n", config_console_logs_broadcast ? "Yes" : "No",config_console_logs_broadcast_port ); strcat( buffer, line );
	sprintf(line, "  Btkcam  Lines   : %d \n",Bcnt); strcat( buffer, line );
	for (ixx=0;ixx<Bcnt;ixx++) { sprintf(line, "  #%d --> %s \n",ixx,Blines[ixx]); strcat( buffer, line ); }
	sprintf(line, "\n"); strcat( buffer, line );
	if ( settings )
		strcpy(settings, buffer);
	else
		printf(buffer);
	
}

void set_debug_level( int level )
{
	config_debug_level = level;
}

void _flog(unsigned int log_transaction, int level, int indent_message, char *message, int indent_data, unsigned char *data, int len, int cols, int newlines_between, int newlines_after, bool show_osd)
{
	int i,j;
	char line[4096];
	
	if( level > config_debug_level ) return;
	
	if( !log_transaction ) _flogm[log_transaction][0] = 0; //if not part of transaction, reset log buffer
	
	//print header
	if( strlen( message ) > 0 )
	{
		for(i=0;i<indent_message;i++) 
		{
			sprintf(line, "  "); strcat( _flogm[log_transaction], line );
		}
		
		sprintf(line, "%s", message); strcat( _flogm[log_transaction], line );
		for(i=0;i<newlines_between;i++) 
		{
			sprintf(line, "\n"); strcat( _flogm[log_transaction], line );
		}
	}
	//print binary data
	if( data != NULL )
	{
		for(i=0;i<len;i++)
		{
			if(cols > 0 )
			{
				if( (i % cols) == 0 )
				{
					if( i > 0 ) 
					{ 
						sprintf(line, "\n"); strcat( _flogm[log_transaction], line );
					}
					if( newlines_between > 0 || strlen( message ) == 0 ) 
						for(j=0;j<indent_data;j++) 
						{
							sprintf(line, "  "); strcat( _flogm[log_transaction], line );
						}
				}
			}
			sprintf(line, "%02X ", data[i]); strcat( _flogm[log_transaction], line );
		}
	}
	for(i=0;i<newlines_after;i++) 
	{
		sprintf(line, "\n"); strcat( _flogm[log_transaction], line );
	}
	if( !log_transaction ) _log_commit_trans( log_transaction, show_osd );
}

void _slog(unsigned int transaction, int level, int indent_message, char *message, int newlines, bool show_osd)
{
	_flog(transaction, level, indent_message, message, 0, NULL, 0, 0, 0, newlines, show_osd);
}

unsigned int _log_begin_trans( void )
{
	pthread_mutex_lock( &_log_transaction_mutex );
	
	for(int i=0;i<4;i++)
	{
		if( (1 << i) & (~_log_mask) )
		{
			_log_mask |= (1 << i);
			_flogm[i + 1][0] = 0; //reset log buffer
			pthread_mutex_unlock( &_log_transaction_mutex );
			return i + 1;
		}
	}
	
	pthread_mutex_unlock( &_log_transaction_mutex );
	return 0;
}

void _log_commit_trans( unsigned int log_trans, bool show_osd )
{
	pthread_mutex_lock( &_log_transaction_mutex );
	pthread_mutex_lock( &_log_display_mutex );

	// -- test log to console
	if( config_log_to_console )
		printf( _flogm[log_trans] );
	
	// -- test log to file
	if( config_log_to_file )
	{
		FILE *logfile = fopen( config_logfile_name, "a" );
		if( logfile != NULL )
		{
			fprintf(logfile, _flogm[log_trans] );
			fclose( logfile );
		}
	}

	// -- test remote broadcast
	if( remote_console_server )
		{
		remote_console_server->broadcast(_flogm[log_trans]);
		}

	
	// -- test show in osd
	//if( show_osd ) box::on_screen_message(_flogm[log_trans]);
	
	if( log_trans ) _log_mask &= ~(1 << (log_trans - 1));
	
	_flogm[log_trans][0] = 0; //reset log buffer
	
	pthread_mutex_unlock( &_log_display_mutex );
	pthread_mutex_unlock( &_log_transaction_mutex );
}

void _log_abort_trans( unsigned int log_trans )
{
	pthread_mutex_lock( &_log_transaction_mutex );
	_log_mask &= ~(1 << (log_trans - 1));
	_flogm[log_trans][0] = 0; //reset log buffer
	pthread_mutex_unlock( &_log_transaction_mutex );
}

void unix2dos(char *dest, char *source)
{
	int length = strlen(source);
	int dest_index = 0;
	
	for(int i=0;i<length;i++)
	{
		if( source[i] == '\n' )
		{
			dest[dest_index++] = '\r';
			dest[dest_index++] = '\n';
		}	
		else dest[dest_index++] = source[i];
	}
	dest[dest_index] = 0;
}


char *_logmf(const char *format, ...)
{
	va_list arg_list;
	
	va_start( arg_list, format );
	vsprintf( _logm, format, arg_list );
	va_end( arg_list );
	
	return _logm;
}

char *get_time( void )
{
	time_t write_time = time(NULL);
	char *string_time = ctime( &write_time );
	
	int length = strlen(string_time);
	if( length > 0 ) string_time[length - 1] = 0;
	
	return string_time;
}

char *get_time_nf( void )
{
	char hms[9] = {0}, day[3] = {0}, month[4] = {0}, year[5] = {0};
	char *temp_time = get_time();
	char *temp_work = new char[strlen(temp_time) + 1];
	strcpy(temp_work, (const char *)temp_time);
	
	strsub(year, temp_work, 20, 4);
	strsub(month, temp_work, 4, 3);
	if( !strcmp("Jan", month) ) strcpy(month, "01");
	if( !strcmp("Feb", month) ) strcpy(month, "02");
	if( !strcmp("Mar", month) ) strcpy(month, "03");
	if( !strcmp("Apr", month) ) strcpy(month, "04");
	if( !strcmp("May", month) ) strcpy(month, "05");
	if( !strcmp("Jun", month) ) strcpy(month, "06");
	if( !strcmp("Jul", month) ) strcpy(month, "07");
	if( !strcmp("Aug", month) ) strcpy(month, "08");
	if( !strcmp("Sep", month) ) strcpy(month, "09");
	if( !strcmp("Oct", month) ) strcpy(month, "10");
	if( !strcmp("Nov", month) ) strcpy(month, "11");
	if( !strcmp("Dec", month) ) strcpy(month, "12");
	strsub(day, temp_work, 8, 2);
	strsub(hms, temp_work, 11, 8);
	
	sprintf(temp_time, "%04d-%02d-%02d|%s", atoi(year), atoi(month), atoi(day), hms);
	
	delete temp_work;
	return temp_time;
}


void 	keystr_to_key(char *src, unsigned char *cw0, unsigned char *cw1)
{
unsigned char 	keys[16];
unsigned char 	*dst = keys;
unsigned char	*end = keys + sizeof(keys);
unsigned int 	tmpu;
int		i,j;
				
dst = keys;
end = keys + sizeof(keys);
while (dst < end && sscanf(src, "%2x", &tmpu) == 1)
	{
	*dst++ = tmpu;
	src += 2;
	}
for (int j=0; j<8; j++) cw0[j]=keys[j];
for (int j=8; j<16; j++) cw1[j-8]=keys[j];
}


void xor_buffers(unsigned char *dest, unsigned char *src_1, unsigned char *src_2, unsigned int length)
{
	//xor two buffers together.
	while(length--) *dest++ = *src_1++ ^ *src_2++;
}

char *strsub(char *dest, char *src, int start, int length)
{
	strncpy(dest, src + start, length);
	dest[length] = 0;
	return dest;
}

char *strltr(char *dest, char *src)
{
	int i, length = strlen(src);
	for(i=0;i<length;i++) if( src[i] != ' ' && src[i] != '\t' ) break;
	strcpy(dest, &src[i]);
	return dest;
}

char *strrtr(char *dest, char *src)
{
	int i, length = strlen(src);
	strcpy(dest, src);
	if( !length ) return dest;
		
	for(i=length-1;i>=0;i--) if( src[i] != ' ' && src[i] != '\t' ) break;
	dest[i + 1] = 0;
	return dest;
}

char *stratr(char *dest, char *src)
{
	char *temp = new char[strlen(src) + 1];
	strltr(temp, src);
	strrtr(dest, temp);
	delete temp;
	return dest;
}

// --------------------------- key and provider configuration management ------------------------------------



int fexists(const char * filename)
{
	if (FILE *file = fopen(filename, "r"))
	{
		fclose(file);
		return 1;
	}
	return 0;
}

int deletefile(const char * filename)
{
return remove(filename);
}

double microseconds( void )
{
	struct timeval start;
	
	gettimeofday(&start, NULL); 
	return (double)start.tv_sec * (double)1000000 + (double)start.tv_usec;
}

int miliseconds( void )
{
	struct timeval start;
	
	gettimeofday(&start, NULL); 
	return (start.tv_sec * 1000000 + start.tv_usec)/1000;
}


int CalcPrio(int type)
{
switch (type)
	{
	case 0x01: return 10;		
	case 0x05: return 10;
	case 0x06: return 1;
	case 0x09: return 5;
	case 0x0b: return 5;
	case 0x17: return 8;
	case 0x18: return 8;
	case 0x0d: return 6;
	default:  return 5;
	}
}


char *file_name_from_path(char *path)
{
	printf("%s\n", path);
	char *current, *last = path;
	while( (current = strstr(last, "/")) ) last = current + 1;
	return last;
}


void dump(unsigned char *buffer, int length)
{
	int i;

	printf("0x0=");
	for (i = 0;i < length;i++)
	{
		if (i % 16 == 0) printf("\n%04X=", i);
		printf("%02X", buffer[i]);
	}
	printf("\n");
}

int getmac( char *mac_addr )
{
	int s;
	struct ifreq buffer;
	char tempchr[3];

	s = socket(PF_INET, SOCK_DGRAM, 0);
	if (s==-1) return -1;

	memset(&buffer, 0x00, sizeof(buffer));
	
	strcpy(buffer.ifr_name, "eth0");

	if (ioctl(s, SIOCGIFHWADDR, &buffer) <0) return -1;

	close(s);

	for( s = 0; s < 6; s++ )
		{
		sprintf(tempchr,"%02X",(unsigned char)buffer.ifr_hwaddr.sa_data[s]);
		memcpy(mac_addr + s*2,tempchr,2);
		}

	return 1;
}

// -------------------------------------------------------------------------------
// ------------------- these are manually implemented as a -----------------------
// ------------- workaround for images that do not include them ------------------

unsigned int _htonl(unsigned int hostlong)
{
	unsigned int netlong = 0;
	((unsigned char *)(&netlong))[0] = (hostlong >> 24) && 0xFF;
	((unsigned char *)(&netlong))[1] = (hostlong >> 16) && 0xFF;
	((unsigned char *)(&netlong))[2] = (hostlong >>  8) && 0xFF;
	((unsigned char *)(&netlong))[3] = (hostlong >>  0) && 0xFF;
	return netlong;
}

unsigned short int _htons(unsigned short int hostshort)
{
	unsigned short int netshort;
	((unsigned char *)(&netshort))[0] = (hostshort >> 8) & 0xFF;
	((unsigned char *)(&netshort))[1] = (hostshort >> 0) & 0xFF;
	return netshort;
}

unsigned int _ntohl(unsigned int netlong)
{
	return ((((unsigned char *)(&netlong))[0]) << 24) | ((((unsigned char *)(&netlong))[1]) << 16) | 
		((((unsigned char *)(&netlong))[2]) << 8) | ((((unsigned char *)(&netlong))[3]) << 0);
}

unsigned short int _ntohs(unsigned short int netshort)
{
	return ((((unsigned char *)(&netshort))[0]) << 8) | ((((unsigned char *)(&netshort))[1]) << 0);
}

void write_ecminfo(char *sentence)
{
FILE * ecmdosya;
ecmdosya=fopen("/tmp/ecm.info","w");
if (ecmdosya!=NULL)
	{
	fputs(sentence,ecmdosya);
	fclose(ecmdosya);
	}
ecmdosya=fopen("/tmp/ecm0.info","w");
if (ecmdosya!=NULL)
	{
	fputs(sentence,ecmdosya);
	fclose(ecmdosya);
	}
}

bool FindKey(char *type, char *val, char *keynr, cEcmInfo *ecm)
{
FILE * keydosya;
char var[4][64];
char var5[2048];
char cline[2048];
keydosya=fopen("/var/keys/SoftCam.Key","r");
if (keydosya==NULL) return false;
memset(ecm->indata,0,0);
//printf("Find Key ->%s  %s  %s \n",type,val,keynr);
while( fgets(cline, 2048, keydosya) )
	{
	config_read_fix_dos_style( cline );
	sscanf(cline,"%s %s %s %s",var[0],var[1],var[2],var5);
	if ( (strcmp(var[0],type)==0) && (strcmp(var[1],val)==0) && (strcmp(var[2],keynr)==0) )
		{
		memcpy(ecm->indata,var5,strlen(var5));
		//printf("Key Found : %s\n",ecm->indata);
		break;
		}
	}
fclose(keydosya);
return false;
}
