#include <pthread.h>        	/* for POSIX threads */
#include <stdio.h>      	/* for printf() and fprintf() */
#include <sys/socket.h> /* for socket(), bind(), and connect() */
#include <arpa/inet.h>  /* for sockaddr_in and inet_ntoa() */
#include <stdlib.h>     /* for atoi() and exit() */
#include <string.h>     /* for memset() */
#include <unistd.h>     /* for close() */
#include <sys/types.h>
#include <sys/time.h>
#include "newcamd/newcamd.h"
#include "newcamd/des.h"
#include "newcamd/newcamd-client.h"
#include "cccam/cccam.h"

#ifdef _CSP
#include "csp/cspmain.h"
void *CSP_runner(void *);		// CSP thread
CSP 		*mycsp=NULL;
bool		CSP_run=false;
pthread_t 	CSPT;
#endif



#define BTKvers		"Vers 2.10  10/01/2013 "

#ifndef MAXLINE
#define MAXLINE 	5			//Max C: & N: line
#endif
#define MAXPROV		25			//max prov caid per Line
#define MAXSID		50			//max SID/CID per ProvCaid
#define MAXDBSZ		512			//max ecm keydb size
#define MAXBISS		200
#define MAXDCW		100
//#define DEBUG 		1

//mynet.c
void DieWithError(char *errorMessage);  /* Error handling function */
int CreateTCPServerSocket(unsigned short port); /* Create TCP server socket */
int AcceptTCPConnection(int servSock);  /* Accept TCP connection request */
int CreateUDPServerSocket(unsigned short port); /* Create TCP server socket */
int check_running_instance(int port);
int network_read(int fd, unsigned char *buffer, unsigned int length, int override_timeout);
int network_write(int fd, unsigned char *buffer, unsigned int length, int override_timeout);
unsigned int GetTickCount(struct timeval initime);
//int __nsleep(const struct timespec *req, struct timespec *rem);
int msleep(unsigned long milisec);

//mycrypt.c
void delspace (char *Str1);
void btkcrypt (char* message, int msglen, char *keyiv );
void btkdecrypt (char* message, int msglen, char *keyiv );
//int BtxtEncry(char *input, char *output);
//int BtxtDecry(char *input, char *output);
//int DCWencode(unsigned char *input, int len);
int DCWdecode(unsigned char *input, int len);


void *ThreadMain(void *arg);            /* Main program of a  TCP thread */
void *ThreadUDP(void *arg);            /* Main program of a UDP thread */
void *Srv_status(void *);		// Status thread


pthread_mutex_t usrmutex     = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t dcwmutex     = PTHREAD_MUTEX_INITIALIZER;			//DCWkeys mutex for ecmsend/zapcam			
pthread_mutex_t keymutex     = PTHREAD_MUTEX_INITIALIZER;			//ECM list mutex - dynamic
pthread_mutex_t nccmutex[MAXLINE];
pthread_mutex_t cccmutex[MAXLINE];

/* Structure of arguments to pass to client threads */

struct ThreadArgs
{
    int clntSock;                      /* Socket descriptor for client */
};


//userlist
struct userlist
	{
	char user[20];
	char pass[20];
	char cver[10];
	int  stat;
	char ip[20];
	bool ok;
	pthread_t tid;
	int ecmtotal;
	int ecmok;
	int ecmnok;
	int ecmcache;
	struct userlist *next;
	};
struct userlist *users,*usr_temp,*usr_cur;	//username pass linked list

struct Routes
	{
	unsigned char kart;
	unsigned char type;
	unsigned char prio;
	};

struct keysdb
	{
	char 	pces[18];
	char 	keys[32];
	//int	onid;
	unsigned char 	tag;		// ecm data 0
	int		chid;		// ird chid
	unsigned int 	ecmcrc;		// correct hash calculated / approved.
	unsigned int 	hashin;		// hash in progress.. for ECM
	unsigned char 	kart,type;
	int		starti;
	};
struct keysdb keysDB[MAXDBSZ];
int dbsize=0;
/*
struct bisskeys
	{
	char pmes[13];
	char keys[33];
	};
bisskeys 	bisslist[MAXBISS];				// BISS & DCW list in dcwbiss.cfg
int		bisscnt=0;
*/
struct dcwkeys
	{
	char chan[20];
	char keys[33];
	time_t zaman;
	};
dcwkeys		dcwlist[MAXDCW];				//DCWs from ecmsend/zapcam/btktrap
int	 	dcwcnt=0;

struct STAT
	{
	unsigned int	ecm;
	unsigned int	ecmok;	
	};

struct LPC
	{
	char 	provcaid[15];		//text in prov caid
	int	sids[MAXSID];
	int	cids[MAXSID];
	int	sidcnt;
	int	cidcnt;
	};

struct NCCs
	{
	char		confn[200];
	struct STAT	stat;
	struct LPC	pc[MAXPROV];			//array of provcaid and sids/cids
	int		pccnt;
	bool		active;
	};
struct 	NCCs	Nlist[MAXLINE];			//ncc liste
int	Ncnt=0;					//ncc count
newcamd_client	*Ncc[MAXLINE];			//newcamd clients


struct CCCs
	{
	char		confn[200];
	struct STAT	stat;
	struct LPC	pc[MAXPROV];
	int		pccnt;
	int		hop;
	bool		active;
	};
struct 	CCCs	Clist[MAXLINE];			//cccam liste
cCardClientCCcam2 *Ccc[MAXLINE];
int	Ccnt=0;					//cccam count

int	dbg=0;						//DEBUG
char	apath[80];					// App  path
int	maxc=1000;					// max client connection
int	conn=0;						// active connection count
char	keyIV[20]= {0};					// Server KEY
char	sport[6]= {0};					// Server Port
char	keyUDP[10]= {0};				// UDP Server KEY
char	uprovcaid[1024];
int	timeout=10000;
int	ECMcfg=1;					// ECM -  0:No ECM          1:Only server Hosted N:& C: (Default)      2:All ECMs
char	Cver[10];					//Client version

int	stat_ecm_total=0;				//Statistics
int	stat_ecm_ok=0;
int 	stat_ecm_nok=0;
int	stat_ecm_cache=0;
int	stat_dcw_ok=0;
int 	stat_dcw_nok=0;

unsigned short echoServPort;     			/* Server port re definition for mynet*/



///MAIN APPLICATION
int main(int argc, char *argv[])
{
int servSock;                    /* Socket descriptor for server */
int clntSock;                    /* Socket descriptor for client */
int stack=0;

pthread_t TCP,UDP,STT;   /* Thread ID from pthread_create() */
struct ThreadArgs *threadArgs;   /* Pointer to argument structure for thread */
FILE * isrtDosya;		// users.txr
char tmpstr[300];		//temp string for users.txt
int  ic,j,iy;
char 	sidchr[5];
int	lastCN=0;		//1 - N: line       2 - C: line



	dbg=0;
	readlink("/proc/self/exe", apath, sizeof(tmpstr));
	ic=strlen(apath)-1;
	while (apath[ic] != '/') ic--;
	apath[ic+1]=0;

 	if (argc >= 2) {
			if (!strncmp(argv[1],"-d",2)) dbg=1;
			if (!strncmp(argv[1],"-C",2)) strcpy(apath,argv[2]); 
			}
	strcpy(tmpstr,apath);
	strcpy(Cver,"070");
	strcat(tmpstr,"dcwserver.cfg\0");
	isrtDosya=fopen(tmpstr,"r");
	if (isrtDosya==NULL)
		{
		fprintf(stderr,"Missing dcwserver.cfg\n");
        	exit(1);
		}
	usr_temp= (struct userlist *)malloc(sizeof(struct userlist));
	users=usr_temp;
	while ( fgets ( tmpstr, 300, isrtDosya ) != NULL )
		{
		if (strncmp(tmpstr,"Debug: ",7)==0) sscanf(tmpstr, "Debug: %d", &dbg);
		if (strncmp(tmpstr,"Port: ",6)==0) sscanf(tmpstr,  "Port: %s", sport);
		if (strncmp(tmpstr,"SKey: ",6)==0) sscanf(tmpstr,  "SKey: %16s", keyIV);
		if (strncmp(tmpstr,"MaxC: ",6)==0)  sscanf(tmpstr, "MaxC: %d", &maxc);	
		if (strncmp(tmpstr,"EcmC: ",6)==0)  sscanf(tmpstr, "EcmC: %d", &ECMcfg);
		if (strncmp(tmpstr,"Tout: ",6)==0)  sscanf(tmpstr, "Tout: %d", &timeout);	
		if (strncmp(tmpstr,"UKey: ",6)==0)  sscanf(tmpstr, "UKey: %s", keyUDP);
		if (strncmp(tmpstr,"Cver: ",6)==0)  sscanf(tmpstr, "Cver: %s", Cver);
		if (strncmp(tmpstr,"Stack: ",7)==0) sscanf(tmpstr, "Stack: %d", &stack);

		if (strncmp(tmpstr,"U: ",3)==0)
			{
			sscanf(tmpstr, "U: %15s %15s %d", usr_temp->user, usr_temp->pass, &usr_temp->stat);
			usr_temp->ok=false; memset(usr_temp->ip,0,sizeof(usr_temp->ip));
			usr_temp->ecmtotal=0;
			usr_temp->ecmok=0;
			usr_temp->ecmnok=0;
			usr_temp->ecmcache=0;
			usr_cur=usr_temp;
			usr_temp= (struct userlist *)malloc(sizeof(struct userlist));
			usr_cur->next=usr_temp;
			}

		if (strncmp(tmpstr,"N: ",3)==0) 
			{ 
			memcpy(Nlist[Ncnt].confn,tmpstr,sizeof(tmpstr)); 
			Nlist[Ncnt].pccnt=0;
			Nlist[Ncnt].active=false;
			Nlist[Ccnt].stat.ecm=0;
			Nlist[Ccnt].stat.ecmok=0;
			Ncnt++; 
			if (Ncnt>MAXLINE) { printf("Only %d N: lines allowed.\n",MAXLINE); exit(0); }
			lastCN=1;    //Last entry is N: line
			}

		if (strncmp(tmpstr,"C: ",3)==0) 
			{ 
			memcpy(Clist[Ccnt].confn,tmpstr,sizeof(tmpstr)); 
			Clist[Ccnt].pccnt=0;
			Clist[Ccnt].active=false;
			Clist[Ccnt].hop=2;
			Clist[Ccnt].stat.ecm=0;
			Clist[Ccnt].stat.ecmok=0;
			Ccnt++; 
			if (Ccnt>MAXLINE) { printf("Only %d C: lines allowed.\n",MAXLINE); exit(0); }
			lastCN=2; 	//Last entry is C: line
			}
		if (strncmp(tmpstr,"Hop: ",5)==0) 
			{
			if (lastCN==2)  sscanf(tmpstr, "Hop: %d",&Clist[Ccnt-1].hop);
			}

		if (strncmp(tmpstr,"ProvCaid: ",5)==0) 
			{
			if (lastCN==1)	
				{
				sscanf(tmpstr, "ProvCaid: %s",Nlist[Ncnt-1].pc[Nlist[Ncnt-1].pccnt].provcaid);
				Nlist[Ncnt-1].pc[Nlist[Ncnt-1].pccnt].sidcnt=0;
				Nlist[Ncnt-1].pc[Nlist[Ncnt-1].pccnt].cidcnt=0;
				Nlist[Ncnt-1].pccnt++;
				}

			if (lastCN==2)  
				{
				sscanf(tmpstr, "ProvCaid: %s",Clist[Ccnt-1].pc[Clist[Ccnt-1].pccnt].provcaid);
				Clist[Ccnt-1].pc[Clist[Ccnt-1].pccnt].sidcnt=0;
				Clist[Ccnt-1].pc[Clist[Ccnt-1].pccnt].cidcnt=0;
				Clist[Ccnt-1].pccnt++;
				}

			}
		if (strncmp(tmpstr,"SID: ",5)==0) 
			{
			delspace(tmpstr);
			if (lastCN==1)	
				for (j=1;j<(strlen(tmpstr)/4);j++)
					{
					int z;
					memcpy(sidchr,tmpstr+(j*4),4);
					sscanf(sidchr,"%4x",&z);
					//printf("%04x\n",z);
					int ni=Ncnt-1;
					int pi=Nlist[Ncnt-1].pccnt-1;
					if (pi<0) pi=0;
					Nlist[ni].pc[pi].sids[Nlist[ni].pc[pi].sidcnt]=z;
					Nlist[ni].pc[pi].sidcnt++;
					}
			if (lastCN==2)
				for (j=1;j<(strlen(tmpstr)/4);j++)
					{
					int z;
					memcpy(sidchr,tmpstr+(j*4),4);
					sscanf(sidchr,"%4x",&z);
					//printf("%04x\n",z);
					int ni=Ccnt-1;
					int pi=Clist[Ccnt-1].pccnt-1;
					if (pi<0) pi=0;
					Clist[ni].pc[pi].sids[Clist[ni].pc[pi].sidcnt]=z;
					Clist[ni].pc[pi].sidcnt++;
					}	
			}
		if (strncmp(tmpstr,"CID: ",5)==0) 
			{
			delspace(tmpstr);
			if (lastCN==1)	
				for (j=1;j<(strlen(tmpstr)/4);j++)
					{
					int z;
					memcpy(sidchr,tmpstr+(j*4),4);
					sscanf(sidchr,"%4x",&z);
					//printf("%04x\n",z);
					int ni=Ncnt-1;
					int pi=Nlist[Ncnt-1].pccnt-1;
					if (pi<0) pi=0;
					Nlist[ni].pc[pi].cids[Nlist[ni].pc[pi].cidcnt]=z;
					Nlist[ni].pc[pi].cidcnt++;
					}
			if (lastCN==2)
				for (j=1;j<(strlen(tmpstr)/4);j++)
					{
					int z;
					memcpy(sidchr,tmpstr+(j*4),4);
					sscanf(sidchr,"%4x",&z);
					//printf("%04x\n",z);
					int ni=Ccnt-1;
					int pi=Clist[Ccnt-1].pccnt-1;
					if (pi<0) pi=0;
					Clist[ni].pc[pi].cids[Clist[ni].pc[pi].cidcnt]=z;
					Clist[ni].pc[pi].cidcnt++;
					}	
			}
		}
	usr_cur->next=NULL;
	free(usr_temp);
	fclose (isrtDosya);
	

	/* parse biss
	strcpy(tmpstr,apath);
	strcat(tmpstr,"biss.cfg\0");
	isrtDosya=fopen(tmpstr,"r");
	if (isrtDosya!=NULL)
		{
		while ( fgets ( tmpstr, 200, isrtDosya ) != NULL )
		   if (strncmp(tmpstr,"DB: ",4)==0) 
			{
			sscanf(tmpstr, "DB: %12s %32s",bisslist[bisscnt].pmes,bisslist[bisscnt].keys );
			bisscnt++;
			}
		fclose (isrtDosya);
		}
	*/

	if (atoi(Cver)<70) strcpy(Cver,"070");

	// Startup of config
	printf("\nStarting Dcwserver %s (%d C/N)... by BillyTK\n",BTKvers,MAXLINE);
	printf("Server Port : %s\n",sport);
	printf("Server Key  : %s\n",keyIV);
	printf("Server Tout : %d\n",timeout);
	printf("Server EcmC : %d\n",ECMcfg);
	printf("UDPsrv Key  : %s\n",keyUDP);
	printf("Client Max# : %d\n",maxc);
	printf("Client Vers : %s\n",Cver);
	printf("Stack  Size : %d\n",stack);
	printf("Server DBG  : %d (%dKB)\n",dbg,sizeof(Nlist)*2);
	

#ifdef DEBUG
	// show users and BISS
	char tempchr[10];
	if (dbg)
		{
		usr_temp=users;
		printf("---------Parsing Users-------\n");
		while (usr_temp!=NULL)
			{
			memset(tmpstr,0,sizeof(tmpstr));
			sprintf(tempchr,"%d",usr_temp->stat);
			strcpy(tmpstr,"User:                Pass:                Stat:      \0");
			memcpy(tmpstr+5,usr_temp->user,strlen(usr_temp->user));
			memcpy(tmpstr+10+16,usr_temp->pass,strlen(usr_temp->pass));
			memcpy(tmpstr+15+32,tempchr,strlen(tempchr));
			printf("%s\n",tmpstr);
			//printf("User/Pass/Stat: %s %s %d \n",usr_temp->user, usr_temp->pass, usr_temp->stat );
			usr_temp=usr_temp->next;
			} 
		/*
		printf("---------BISS Entries--------\n");
		for (ic=0;ic<bisscnt;ic++)
			printf("BISS %02d: %s:%s\n",ic+1,bisslist[ic].pmes,bisslist[ic].keys);
		*/
		}
#else
#warning No DEBUG defined in compilation...
#endif	
	
	// Create NCC clients
	for(int i=0;i<Ncnt;i++)
    		{
		printf("\n----------Newcamd Lines--------------\n");
		if (Ncc[i] = new newcamd_client(Nlist[i].confn,timeout,VERSION_525), !Ncc[i])
			{
			printf("Newcamd Client [%d] init Failed.",i);
			exit(0);
			}
		Ncc[i]->dbg=dbg;
		Nlist[i].active=Ncc[i]->is_ready();
		if (Nlist[i].pccnt==0) 
			{
			for (j=0 ;j<Ncc[i]->card_provider_id_count;j++) 
				sprintf(Nlist[i].pc[j].provcaid,"%06X%04X",Ncc[i]->card_provider_id[j],Ncc[i]->card_caid);
			Nlist[i].pccnt=Ncc[i]->card_provider_id_count;
			}
		}
	//printf("%d Newcamd Clients Created.\n",Ncnt);
	// Create NCC mutexes
	for(int i=0;i<Ncnt;i++) pthread_mutex_init(&nccmutex[i], NULL);
	//printf("%d Ncc Mutex init done.\n\n",Ncnt);

	
	// Create CCCam  Clients
	for(int i=0;i<Ccnt;i++)
		{
		printf("\n----------CCcam Lines------------\n");
		if (Ccc[i] = new cCardClientCCcam2(Clist[i].confn,timeout,Clist[i].hop), !Ccc[i])
			{
			printf("CCcam Client [%d] init Failed.",i);
			exit(0);
			}
		Clist[i].active=Ccc[i]->ready;
		Ccc[i]->dbg = dbg;
		}
	//printf("%d CCcam Clients Created.\n",Ccnt);
	// Create NCC mutexes
	for(int i=0;i<Ccnt;i++) pthread_mutex_init(&cccmutex[i], NULL);
	//printf("%d Ccc Mutex init done.\n\n",Ccnt);

	
	//Prepare user-provcaid list
	memset(uprovcaid,0,sizeof(uprovcaid));
	//printf("---- N: Lines ----\n");
	for(int i=0;i<Ncnt;i++)  
		for (int j=0;j<Nlist[i].pccnt;j++)
			{
			printf("Line N[%d]->  ProvCaid: %s   Sid Cnt:%d  ChID Cnt:%d\n",i,Nlist[i].pc[j].provcaid,Nlist[i].pc[j].sidcnt,Nlist[i].pc[j].cidcnt);
			for(int ic=0;ic<Nlist[i].pc[j].sidcnt;ic++) 
				{
				printf("%04x",Nlist[i].pc[j].sids[ic]);
				if (Nlist[i].pc[j].cidcnt>0) printf("(%04x)",Nlist[i].pc[j].cids[ic]);
				printf(" ");
				}
			if (Nlist[i].pc[j].sidcnt) printf("\n");

			if (!memcmp(Nlist[i].pc[j].provcaid,"FFFFFFFFFFFF",10)) ECMcfg=2;
			else	{
				for(iy=0;iy<strlen(uprovcaid)/10;iy++) 
					if (!memcmp(uprovcaid+iy*10,Nlist[i].pc[j].provcaid,10)) break;
				if (iy==strlen(uprovcaid)/10)
					strcat(uprovcaid,Nlist[i].pc[j].provcaid);
				}
			}

	//printf("---- C: Lines ----\n");
	for(int i=0;i<Ccnt;i++)  
		if (Clist[i].pccnt==0 && ECMcfg==2) printf("Line C[%d]->  ProvCaid: FFFFFFFFFF   Sid#0 :\n",i);
		else if (Clist[i].pccnt==0 && ECMcfg==1) printf("Line C[%d]->  Disabled.  No Routing. \n",i);
		else for (int j=0;j<Clist[i].pccnt;j++)
			{
			printf("Line C[%d]->  ProvCaid: %s   Sid Cnt:%d  ChID Cnt:%d\n",i,Clist[i].pc[j].provcaid,Clist[i].pc[j].sidcnt,Clist[i].pc[j].cidcnt);
			for(int ic=0;ic<Clist[i].pc[j].sidcnt;ic++) 
				{
				printf("%04x",Clist[i].pc[j].sids[ic]);
				if (Clist[i].pc[j].cidcnt>0) printf("(%04x)",Clist[i].pc[j].cids[ic]);
				printf(" ");
				}
			if (Clist[i].pc[j].sidcnt)  printf("\n");

			if (!memcmp(Clist[i].pc[j].provcaid,"FFFFFFFFFFFF",10)) ECMcfg=2;
			else	{
				for(iy=0;iy<strlen(uprovcaid)/10;iy++) 
					if (!memcmp(uprovcaid+iy*10,Clist[i].pc[j].provcaid,10)) break;
				if (iy==strlen(uprovcaid)/10)
					strcat(uprovcaid,Clist[i].pc[j].provcaid);
				}
			}
	
	if (strlen(uprovcaid)/10>100) 
		{ 
		printf("Maximum 100 providers can be supported for all N: & C: lines\nWarning Setting EcmC: 2 and deleting ProvCaids\n"); 
		strcpy(uprovcaid,"FFFFFFFFFF");
		}
	//printf("%d Client ProvCaid: %s\n",strlen(uprovcaid)/10,uprovcaid);

	echoServPort = atoi(sport);

#ifdef _CSP
	///---------CSP thread activation---------------------
	if (mycsp = new CSP(apath), !mycsp)	{ perror("CSP init Failed.");  exit(0); }
	if (!mycsp->ready) 
		{
		printf("CSP-cache has been disabled.\n");
		delete(mycsp);
		mycsp=NULL;
		}
		else 
		{	
		printf("CSP-cache has been started.\n");
		if (pthread_create(&CSPT, NULL, CSP_runner, NULL) != 0) DieWithError("pthread_create() failed");
		printf("CSP-Cache Server thread created.\n");
		msleep(1000);
		}
#endif

	///---------UDP thread activation---------------------
	if (strlen(keyUDP)>0)
		{
        	if (pthread_create(&UDP, NULL, ThreadUDP, NULL) != 0) DieWithError("pthread_create() failed");
		printf("UDPKey Server thread created.\n");
		}

	//----------Create System Status Thread------------.  //restores down NCCs
	if (pthread_create(&STT, NULL, Srv_status, NULL) != 0) DieWithError("pthread_create() failed");
		printf("System Status Thread Created.\n");

    	//---------------TCP Client Thread creation in loop -----------------------
	servSock = CreateTCPServerSocket(echoServPort);

    	while (1) 
    		{
		Loop:
		clntSock = AcceptTCPConnection(servSock);

		if (conn>=maxc)    // strcmp(clnt_ip1,"83.86.205.69")==0) ||
			{
			shutdown(clntSock,2);    
			close (clntSock);
			goto Loop;
			}
			

        	/* Create separate memory for client argument */
        	if ((threadArgs = (struct ThreadArgs *) malloc(sizeof(struct ThreadArgs))) == NULL)
        	 	DieWithError("malloc() failed");
        	threadArgs -> clntSock = clntSock;
		if (stack!=0)
			{
			pthread_attr_t attr;
			pthread_attr_init(&attr);
			pthread_attr_setstacksize(&attr, stack*1024);
        		/* Create client thread */
        		if (pthread_create(&TCP, &attr, ThreadMain, (void *) threadArgs) != 0)
            			DieWithError("pthread_create() failed");
			pthread_attr_destroy(&attr);
			}
		else if (pthread_create(&TCP, NULL, ThreadMain, (void *) threadArgs) != 0)
            			DieWithError("pthread_create() failed");
		//sleep(1);
    		}
    /* NOT REACHED */
}



int check_ECMcache(char *cpces, unsigned int hash)	
{	
	int  	location=-1;
	int 	j;
	for (j=0;j<dbsize;j++)
		if ( (memcmp(cpces,keysDB[j].pces,18)==0) )
			{ 
			if (keysDB[j].ecmcrc==hash) location=j;
				else location=100000+j;
			break; 
			}
	if (j==dbsize) location=-1;
	return location;
}

int addto_ECMcache(int dbcurs, char *cpces, unsigned char ecm0, int chid, unsigned int hash, char *keys,unsigned char kart,unsigned char type,int starti)
{
	pthread_mutex_lock( &keymutex );
	if (dbcurs>=100000) dbcurs-=100000;
	else if (dbcurs==-1)
		{
		dbcurs=dbsize;
		dbsize++;
		if (dbsize>=MAXDBSZ) { dbcurs=0; dbsize=1; };
		memcpy(keysDB[dbcurs].pces,cpces,18);
		}
	memcpy(keysDB[dbcurs].keys,keys,32);
	keysDB[dbcurs].tag=ecm0;
	keysDB[dbcurs].chid=chid;
	keysDB[dbcurs].ecmcrc=hash;
	keysDB[dbcurs].hashin=hash;
	keysDB[dbcurs].kart=kart;
	keysDB[dbcurs].type=type;
	keysDB[dbcurs].starti=starti;
	if (dbg) printf("Key inserted to DB[%d/%d]  -> Kart:%d  Type:%d  Si:%d\n",dbcurs+1,dbsize,kart,type,starti);
	pthread_mutex_unlock( &keymutex );
	return dbcurs;
}


void *ThreadMain(void *threadArgs)
{
	int clntSock;                   /* Socket descriptor for client connection */
	char echoBuffer[1024];          /* Buffer for echo string */
	unsigned char *echo= (unsigned char*)echoBuffer;
	int recvMsgSize;                    /* Size of received message */
	char cmd1[5];
	char cmd2[64];
	unsigned char ecmdata[512];
	int ecmlen;
	bool auth, keyok;
	struct userlist *srcuser=NULL;
	
	sockaddr_in claddr;
	socklen_t clen;
	char clnt_ip[20];

	int 		j;
	int		cmd_type;
	int 		dbcur,chidx;
	unsigned int	provx,crcx;
	unsigned short  caidx,sidx,pidx;
	char		prov1caid1[11];

	struct Routes 	rlist[MAXLINE*2];

	struct Routes 	rslist[MAXLINE*2];
	int		rcnt,rxcnt=0;
	int 		Kart,starti;
	bool		fastroute;

	time_t		simdi;
	cEcmInfo 	ecmx;

#ifdef _CSP		
	//CSP Variables
	struct 	timeval startime;
	struct 	cache_data cspcache;
	struct 	cache_data *cspreply;
#endif


	conn++;
	auth=false;
	/* Guarantees that thread resources are deallocated upon return */
	pthread_detach(pthread_self()); 
	/* Extract socket file descriptor from argument */
	clntSock = ((struct ThreadArgs *) threadArgs) -> clntSock;
	free(threadArgs);              /* Deallocate memory for argument */
	clen = sizeof(claddr);
	getpeername(clntSock, (sockaddr*)&claddr, &clen);
	strcpy(clnt_ip,inet_ntoa(claddr.sin_addr));
	
	if (dbg) printf("Connection From     : %s\n",clnt_ip);
	//if (conn>maxc) goto GEBER;
	//send welcome message and start loop
	memset(echoBuffer,0,sizeof(echoBuffer));
	strcpy(echoBuffer,"Welcome to DcwServer by BillyTK\0\0");
	if (network_write(clntSock,echo, 32,timeout) != 32)  goto GEBER;

	/* Receive message from client for authtenication*/
	memset(echoBuffer,0,sizeof(echoBuffer));
	recvMsgSize = network_read(clntSock, echo, 1024,timeout);
	if ( recvMsgSize > 60 || recvMsgSize < 9) goto GEBER;
	btkdecrypt(echoBuffer,recvMsgSize,keyIV);
	recvMsgSize -= 2;
	memset(cmd1,0,sizeof(cmd1));
	memset(cmd2,0,sizeof(cmd2));
	memcpy(cmd1,echoBuffer,4);
	memcpy(cmd2,echoBuffer+4,recvMsgSize-4);
	//if (dbg) printf("Received Command:%s\n",echoBuffer);

	if (strncmp(cmd1,"USER",4)==0)
		{
		pthread_mutex_lock( &usrmutex );
		srcuser=users;
		while ( (srcuser!=NULL) && (auth==false) )
			{
			strcpy(cmd2,srcuser->user);strcat(cmd2,srcuser->pass);
			//if (dbg) printf("Users ?? %s == %s\n",cmd2,cmdtmp);
			if ((memcmp(cmd2,echoBuffer+4,recvMsgSize-7)==0)  && (srcuser->stat!=0) ) 
				{
				auth=true;
				memset(srcuser->cver,0,sizeof(srcuser->cver));
				memcpy(srcuser->cver,echoBuffer+recvMsgSize-3,3);
				}
			 else srcuser=srcuser->next;
			}
		if (auth)
			{
			if (srcuser->stat==1 && srcuser->ok)  { if (dbg) printf("Double User Activity: %s\n",srcuser->user); auth=false; }
			if (atoi(srcuser->cver)<atoi(Cver))   { if (dbg) printf("Server[%s] / Client[%d] Miss-match\n",Cver,atoi(srcuser->cver)); auth=false; }
			}
		pthread_mutex_unlock( &usrmutex );	
		if (!auth) goto GEBER;

		memset(echoBuffer,0,sizeof(echoBuffer));
 		strcpy(echoBuffer,"ASK:");
		strcat(echoBuffer,Cver); 	
		if (ECMcfg> 0) strcat(echoBuffer,uprovcaid);
		if (ECMcfg==2) strcat(echoBuffer,"FFFFFFFFFF\0");
		//printf("Echobuffer %d : %s\n",strlen(uprovcaid),echoBuffer);
		recvMsgSize=strlen(echoBuffer);
		btkcrypt(echoBuffer,recvMsgSize,keyIV);
		recvMsgSize += 2;
		if (network_write(clntSock, echo,recvMsgSize,timeout ) != recvMsgSize )  goto GEBER;
		}
	if (auth)
		{
		pthread_mutex_lock( &usrmutex );
		srcuser->ok=true;
		strcpy(srcuser->ip,clnt_ip);
		srcuser->tid=pthread_self();
		pthread_mutex_unlock( &usrmutex );
		if (dbg) printf ("User:%s  Pass:%s  IP:%s  Ver:%s TID:%ld\n",srcuser->user,srcuser->pass,srcuser->ip,srcuser->cver,(long int) srcuser->tid  );
		rcnt=starti=0;
		fastroute=false;
		}
	//Start Loop after authentication
	while (auth)
		{
		/* Receive message from client  after authtenication*/
		memset(echoBuffer,0,sizeof(echoBuffer));
	
		recvMsgSize = network_read(clntSock,echo, 1000,timeout*24);
		if (recvMsgSize< 7) 
				{ 
				if (dbg) printf("Invalid message size (<7) : %d\n",recvMsgSize);
				goto GEBER;
				}
		
		if ( recvMsgSize > 500) 
				{ 
				if (dbg) printf("Invalid message size (>500) : %d\n",recvMsgSize);
				goto GEBER;
				}
		btkdecrypt(echoBuffer,recvMsgSize,keyIV);
		recvMsgSize -= 2;
		memset(cmd1,0,sizeof(cmd1));
		memset(cmd2,0,sizeof(cmd2));
		memcpy(cmd1,echoBuffer,4);
		keyok=false;
		cmd_type=0;
		
		if (recvMsgSize==8) goto Respond1;			 

		//if (dbg)  printf("-----------------------------------\n"); 
			
		if (memcmp(cmd1,"ASK:",4)==0)
			{
			cmd_type=1;
			rcnt=starti=0; fastroute=false;	//reset ECM parameters
			memcpy(cmd2,echoBuffer+4,recvMsgSize-4);
			if (dbg) printf("\nReceived Command    : %s%s (%d) from %s %s\n",cmd1,cmd2,recvMsgSize,clnt_ip,srcuser->user);
			if ( (dcwcnt>0) && (recvMsgSize==20) )
				{
				//if (dbg) printf("Searching DCW   Key : %s\n",cmd2);
				pthread_mutex_lock( &dcwmutex );
				j=0;
				time (&simdi);
				while ((!keyok) && (j<dcwcnt))
				     if (memcmp(dcwlist[j].chan,cmd2,16)==0  && difftime(simdi,dcwlist[j].zaman)<(timeout*1.7/1000) ) keyok=true; 
						else j++; 
				if (keyok) {
					strcpy(echoBuffer,"KEY:"); 
					memcpy(echoBuffer+4,dcwlist[j].keys,32);
					echoBuffer[36]=0;
					stat_dcw_ok++;
					}
				pthread_mutex_unlock( &dcwmutex );
				}
			goto Respond1;
			}	

		if ( (memcmp(cmd1,"ECM:",4)==0) && (recvMsgSize>30) && ( (Ncnt>0) || (Ccnt>0) ) && (ECMcfg>0) ) // Basic ECM command Variables set
			{
			cmd_type=2;
			memcpy(cmd2,echoBuffer+4,18);			//ECM: PROVID  CAID PID0 SID0 + ecmdata
			memcpy(ecmdata,echoBuffer+22,recvMsgSize-22);
			ecmlen=recvMsgSize-22;
			if (dbg) printf("\nReceived Command    : %s%s (%d) from %s %s\n",cmd1,cmd2,recvMsgSize,clnt_ip,srcuser->user);

			memset(prov1caid1,0,11); memcpy(prov1caid1,cmd2,6);	sscanf(prov1caid1 ,"%6X",&j); provx=j;
			memset(prov1caid1,0,11); memcpy(prov1caid1,cmd2+6,4);	sscanf(prov1caid1 ,"%4X",&j); caidx=j;
			memset(prov1caid1,0,11); memcpy(prov1caid1,cmd2+10,4);	sscanf(prov1caid1 ,"%4X",&j); pidx=j;
			memset(prov1caid1,0,11); memcpy(prov1caid1,cmd2+14,4);	sscanf(prov1caid1 ,"%4X",&j); sidx=j;
			
			if (dbg)  printf("Prov:%06X  Caid:%04X  Ecm:%04X  Sid:%04X \n",provx,caidx,pidx,sidx);
			
			// beta tunnel
			if ( ((caidx==0x1833) || (caidx==0x1834))  &&  ((ecmlen >= 134) && (ecmlen < 144)) )
				{
				if (dbg) printf("Beta tunnel : %04X\n",caidx);
				unsigned char headerN3[10] =       {0xc7, 0x00, 0x00, 0x00, 0x01, 0x10, 0x10, 0x00, 0x87, 0x12};
				unsigned char headerN2[10] =       {0xc9, 0x00, 0x00, 0x00, 0x01, 0x10, 0x10, 0x00, 0x48, 0x12};

				if (caidx==0x1833) caidx=0x1702; else caidx=0x1722;
				provx=0;
				memmove(ecmdata+13, ecmdata+3, ecmlen-3);
				if (ecmlen>0x88)  
					{
					memcpy(ecmdata+3, headerN3, 10);
					if (ecmdata[0]==0x81) ecmdata[12]+= 1;
					ecmdata[1]=0x70;
					}
				   else memcpy(ecmdata+3, headerN2, 10);
				ecmlen+=10;
				ecmdata[2]=ecmlen-3;
				}
			if (caidx == 0x1702 && ecmlen == 0x89 && ecmdata[3] == 0x07 && ecmdata[4] == 0x84) caidx=0x1833;

			chidx=0;
			if ((caidx & 0xff00)==0x600) chidx=(ecmdata[6]<<8)+ecmdata[7];
			crcx=0;
			for (j=0;j<ecmlen-3;j++) crcx = 31*crcx + ecmdata[j+3];  //crcx+=ecmdata[j];	//crc of ecm
			ecmx.caId=caidx;  ecmx.provId=provx;		// ecm struct for cccam
			ecmx.prgId=sidx;  ecmx.ecm_pid=pidx;
			memset(prov1caid1,0,11);  sprintf(prov1caid1,"%06X%04X",provx,caidx);
#ifdef _CSP			
			gettimeofday( &startime, NULL );
			if (mycsp) 
				{
				cspcache.tag=ecmdata[0];
				cspcache.sid=sidx; cspcache.prov=provx;
				cspcache.onid=0;   cspcache.caid=caidx;
				cspcache.hash=crcx;
				cspreply=mycsp->BTK_to_CSP(&cspcache);
				if (cspreply!=NULL) goto Respond1;
				}
#endif
			} else goto Respond1;
//FIND ROUTES 
		if (rcnt==0) 
			{
			//Find CCcam Routes:
			fastroute=false;
			for (int cl=0;cl<Ccnt;cl++) if (Clist[cl].active)
				{
				if (Clist[cl].pccnt==0 && ECMcfg==2) 			//No ProvCaid & EcmCfg=2
					{
					rlist[rcnt].type=1;				//cccam
					rlist[rcnt].kart=cl;
					rlist[rcnt].prio=0;				// 0 -  FFFFFFF or ecmcfg=2
					rcnt++; 
					continue;
					}
				for (j=0;j<Clist[cl].pccnt;j++)
					{
					if (memcmp(Clist[cl].pc[j].provcaid,"FFFFFFFFFFFF",10)==0)	//
						{
						rlist[rcnt].type=1;				//cccam
						rlist[rcnt].kart=cl;
						rlist[rcnt].prio=0;				// 0 -FFFFFFF or ecmcfg=2
						rcnt++; 
						break;		//bj
						}
						if (memcmp(Clist[cl].pc[j].provcaid,prov1caid1,10)==0)
						{
						if (Clist[cl].pc[j].sidcnt==0)
							{
							rlist[rcnt].type=1;			//cccam
							rlist[rcnt].kart=cl;
							rlist[rcnt].prio=1;			// 1- only provcaid
							rcnt++; 
							break;		//bj
							}
						for (int oo=0;oo<Clist[cl].pc[j].sidcnt;oo++)
							if (Clist[cl].pc[j].sids[oo]==sidx) 
								{
								if  (Clist[cl].pc[j].cidcnt==0)
									{
									rlist[rcnt].type=1;			//cccam
									rlist[rcnt].kart=cl;
									rlist[rcnt].prio=2;			// 2-  provcaid + SID
									rcnt++; 
									break;		//boo
									}
								if (Clist[cl].pc[j].cidcnt>0 && ((caidx>>8)==6) )
									if ( Clist[cl].pc[j].cids[oo] == (ecmdata[6]<<8) + ecmdata[7] )
										{
										rlist[rcnt].type=1;			//cccam
										rlist[rcnt].kart=cl;
										rlist[rcnt].prio=3;			// 3-  provcaid + SID + CID
										rcnt++; 
										break;	//boo
										}
								break;	//boo
								}
						break;	//bj
						}
					} //loop provcaid of C: line
				} //loop C: line count
			
			//Find Newcamd routes
			for (int nl=0;nl<Ncnt;nl++) if (Nlist[nl].active)
				for (j=0;j<Nlist[nl].pccnt;j++)
					if (memcmp(Nlist[nl].pc[j].provcaid,prov1caid1,10)==0)
						{
						if (Nlist[nl].pc[j].sidcnt==0)
							{
							rlist[rcnt].type=2;				//newcamd
							rlist[rcnt].kart=nl;
							rlist[rcnt].prio=1;				// 1- only provcaid
							rcnt++; 
							break;	//bj
							}	
						for (int oo=0;oo<Nlist[nl].pc[j].sidcnt;oo++)
							if (Nlist[nl].pc[j].sids[oo]==sidx) 
								{
								if (Nlist[nl].pc[j].cidcnt==0)
									{
									rlist[rcnt].type=2;				//newcamd
									rlist[rcnt].kart=nl;
									rlist[rcnt].prio=2;				// 2- Provcaid + SID
									rcnt++; 
									break;  //boo
									}
								if (Nlist[nl].pc[j].cidcnt>0 && (caidx>>8)==6)
									if ( Nlist[nl].pc[j].cids[oo] == (ecmdata[6]<<8) + ecmdata[7] ) break;
										{
										rlist[rcnt].type=2;			//newcamd
										rlist[rcnt].kart=nl;
										rlist[rcnt].prio=3;			// 3-  provcaid + SID + CID
										rcnt++; 
										break;	//boo
										}
								break;	//boo
								}
						break;   //bj
						}
			if (rcnt>0) 
				{
				//if (dbg) printf("Printing %d Routes.\n",rcnt);
				rxcnt=0;
				for (int px=3;px!=-1;px--)			//prio-->  3-0
					for (int tx=2;tx>0;tx--)		//type-->  N:2    C:1   -----B:0 yok
						for (int rx=0;rx<rcnt;rx++)	//routes
					    	if (rlist[rx].prio==px  && rlist[rx].type==tx)
							{
							rslist[rxcnt].prio=rlist[rx].prio;
							rslist[rxcnt].type=rlist[rx].type;
							rslist[rxcnt].kart=rlist[rx].kart;
							if (dbg) printf("Route[%d] -> Kart:%d  Type:%d  Prio:%d\n",
									rxcnt,rslist[rxcnt].kart,rslist[rxcnt].type,rslist[rxcnt].prio);
							rxcnt++;
							}
				rxcnt=-1; //initialize
				} 
			else 
				{
				if (dbg) printf("No Route Found.\n");
				goto Respond1;
				}
				
			}//  End of find routes
//CheckCache:
		//cache  route
		pthread_mutex_lock( &keymutex );
		j=0;
		dbcur=check_ECMcache(cmd2,crcx);	// dbcur -> 0-1000: dcw ok     100000+ : pces ok ecmhash not       -1: not in cache
		if (dbcur>=100000) 
			{
			if ( (chidx) && (keysDB[dbcur-100000].chid!=chidx) ) 
				{ 
				printf("Not valid Chid (%X) for Irdeto.\n",chidx);
				pthread_mutex_unlock( &keymutex );
				goto Respond1; 
				}
			if (keysDB[dbcur-100000].hashin==crcx) 
				{   				     
				if (dbg) printf("Waiting in ECMCache : DBcur %d\n",dbcur-100000); 
				for (j=1;j<30;j++)
					{
					pthread_mutex_unlock( &keymutex );
					msleep(100);
					pthread_mutex_lock( &keymutex );
					if (keysDB[dbcur-100000].ecmcrc==crcx) { dbcur=dbcur-100000; break; }
					}
				}
			else keysDB[dbcur-100000].hashin=crcx;	
			}
		if (dbcur>=0 && dbcur<100000)
			{
			memset(echoBuffer,0,sizeof(echoBuffer));
			strcpy(echoBuffer,"KEY:"); 
			memcpy(echoBuffer+4,keysDB[dbcur].keys,32);
			keyok=true;
			stat_ecm_cache++;
			srcuser->ecmcache++;
			if (dbg) printf("KEY Found in P-Cache: DBcur %d\n",dbcur);
			// fix fastroute
			fastroute=true;
			rcnt=1; 
			rxcnt=0;
			rslist[0].kart=keysDB[dbcur].kart;
			rslist[0].type=keysDB[dbcur].type;
			starti=keysDB[dbcur].starti;
			j=1;
			} 
		pthread_mutex_unlock( &keymutex );
		if (j>=1)		// ecm wait happened
			{
			if (!keyok) if (dbg) printf("BtkCache Wait Failed with No DCW.\n");
			goto Respond1; 
			}
		if (fastroute && rcnt>0)
			{
			Kart=rslist[rxcnt].kart;
			if (dbg) printf("Fast Routing -> Type:%d  Kart:%d  Si:%d  (%d)\n",rslist[rxcnt].type,rslist[rxcnt].kart,starti,rcnt);
			if (rslist[rxcnt].type == 1) goto ForwardCCC;
						else goto ForwardNCC;
			}

TryRoute:
		rxcnt++;
		if (rxcnt>=rcnt) goto Respond1; 
		Kart=rslist[rxcnt].kart;
		starti=0;
		if (dbg) printf("Try Next Route -> %d of %d\n",rxcnt,rcnt);
		if (rslist[rxcnt].type == 1) goto ForwardCCC;
					else goto ForwardNCC;
		
ForwardNCC:
		//if (dbg) printf("Trying N[%d] for %04X\n",Kart,sidx);
		// start ncc processing
		pthread_mutex_lock( &nccmutex[Kart]);
		if (!Ncc[Kart]->is_ready())
			{
			Nlist[Kart].active=false;
			pthread_mutex_unlock( &nccmutex[Kart]);
			goto TryRoute;
			}
		Ncc[Kart]->send_ecm(ecmdata,ecmlen,sidx);
		Nlist[Kart].stat.ecm++;
		if (!Ncc[Kart]->status) 
			{
			char 	mycw[3];
			memset(echoBuffer,0,sizeof(echoBuffer));
			strcpy(echoBuffer,"KEY:");
			for (j=0;j<8;j++) 
				{
				sprintf(mycw,"%02X",Ncc[Kart]->cw_0[j]);
				strcat(echoBuffer,mycw);
				}
			for (j=0;j<8;j++) 
				{
				sprintf(mycw,"%02X",Ncc[Kart]->cw_1[j]);
				strcat(echoBuffer,mycw);
				}
			keyok=true;
			addto_ECMcache(dbcur,cmd2,ecmdata[0],chidx,crcx,echoBuffer+4,rslist[rxcnt].kart,rslist[rxcnt].type,0);
			Nlist[Kart].stat.ecmok++;
			fastroute=true;
			}
		Nlist[Kart].active=Ncc[Kart]->is_ready();
		pthread_mutex_unlock( &nccmutex[Kart] );
		//end of  NCC ECM command
		if (!keyok) goto TryRoute; else goto Respond1;
			

ForwardCCC:
		//if (dbg) printf("Trying C[%d] for %04X\n",Kart,sidx);
		unsigned char cw16[16];
		pthread_mutex_lock( &cccmutex[Kart] );
		if (!Ccc[Kart]->ready)
			{
			Clist[Kart].active=false;
			pthread_mutex_unlock( &cccmutex[Kart] );
			goto TryRoute;
			}
		
		while ( (starti=Ccc[Kart]->HasECM(&ecmx,starti)) >= 0 ) 
			{
			Clist[Kart].stat.ecm++;
			if (Ccc[Kart]->ProcessECM(&ecmx,ecmdata,cw16) )
				{
				char 	mycw[3];
				memset(echoBuffer,0,sizeof(echoBuffer));
				strcpy(echoBuffer,"KEY:");
				for (j=0;j<16;j++) 
					{
					sprintf(mycw,"%02X",cw16[j]);
					strcat(echoBuffer,mycw);
					}
				keyok=true;
				starti-=1;
				addto_ECMcache(dbcur,cmd2,ecmdata[0],chidx,crcx,echoBuffer+4,rslist[rxcnt].kart,rslist[rxcnt].type,starti);
				Clist[Kart].stat.ecmok++;
				fastroute=true;
				break;
				} 
			}
		Clist[Kart].active=Ccc[Kart]->ready;		
		pthread_mutex_unlock( &cccmutex[Kart] );	
		//end ECM CCCC
		if (!keyok) goto TryRoute;


Respond1:
		if  (!keyok)  
			{
			rcnt=0;
			starti=0;
			fastroute=false;
			strcpy(echoBuffer,"ASK:NOP\0");
			}
#ifdef _CSP
		if (mycsp && !keyok && (cmd_type==2))
			{	
			int init1=0;
			cspreply=mycsp->BTK_to_CSP(&cspcache);
			if (cspreply==NULL )
				{
				init1=mycsp->cspwait-GetTickCount(startime);
				if (init1>0) 
					{ 
					if (dbg) printf("Sleeping  %d msec for CSP cache...\n",init1);  
					for (j=0;j<init1;j+=200)
						{
						msleep(200);
						cspreply=mycsp->BTK_to_CSP(&cspcache);
						if (cspreply!=NULL) break;
						}
					init1=j;
					}
				}
			if (cspreply!=NULL )
					{
					char 	mycw[3];
					memset(echoBuffer,0,sizeof(echoBuffer));
					strcpy(echoBuffer,"KEY:");
					for (j=0;j<16;j++) 
						{
						sprintf(mycw,"%02X",cspreply->cw[j]);
						strcat(echoBuffer,mycw);
						}
					//if (dbg) 
					stat_ecm_cache++;
					srcuser->ecmcache++;
					if (dbg) printf("KEY Found in C-Cache: %d msec\n",init1);
					keyok=true;
					}			
			}
#endif	

		if (cmd_type==2)
			{
			stat_ecm_total++;
			if (keyok) stat_ecm_ok++; else stat_ecm_nok++;
			srcuser->ecmtotal++;
			if (keyok) srcuser->ecmok++; else srcuser->ecmnok++;
			} 
		if (dbg & (recvMsgSize!=8)) printf("Command  Response   : %s to %s\n\n",echoBuffer,clnt_ip);
		recvMsgSize=strlen(echoBuffer);
		if ( recvMsgSize > 60 || recvMsgSize < 7 ) goto GEBER;
		btkcrypt(echoBuffer,recvMsgSize,keyIV);
		recvMsgSize += 2;

		if (network_write(clntSock, echo,recvMsgSize,timeout ) != recvMsgSize ) 
				{ 
				if (dbg) printf("Error in Write message : %d\n",recvMsgSize);
				goto GEBER;
				}
		if ((pthread_self()!=srcuser->tid) && (srcuser->stat==1)) auth=false;
		}
	
	GEBER:
	shutdown(clntSock,2);    /* Close client socket */
	close (clntSock);
	if (dbg) printf("Client disconnected : %s\n",clnt_ip);
	
	pthread_mutex_lock( &usrmutex );
	if (auth) srcuser->ok=false;
	conn--;
	pthread_mutex_unlock( &usrmutex );
		
	pthread_exit ((void *)0);
}


////////////////////////////////////////////////

void *ThreadUDP(void *threadArgs)
{
int  udptsock;
struct sockaddr_in udpClntAddr; /* Client address */
unsigned int cliAddrLen;         /* Length of incoming message */
char 	udpBuffer[255];        /* Buffer for echo string */
int 	recvMsgSize;                 /* Size of received message */
int 	ix;
int 	curdcw;
time_t  simdi1;

    /* Guarantees that thread resources are deallocated upon return */
    pthread_detach(pthread_self()); 

    /* Extract socket file descriptor from argument */
    udptsock = CreateUDPServerSocket(echoServPort);

    for (;;) /* Run forever */
    	{
        /* Set the size of the in-out parameter */
        cliAddrLen = sizeof(udpClntAddr);
	memset(udpBuffer,0,sizeof(udpBuffer));
        /* Block until receive message from a client */
        if ((recvMsgSize = recvfrom(udptsock, udpBuffer, 100, 0, (struct sockaddr *) &udpClntAddr, &cliAddrLen)) < 0)
            DieWithError("UDPthread recvfrom() failed");
	DCWdecode((unsigned char*)udpBuffer,recvMsgSize);

	if ( (memcmp(udpBuffer+recvMsgSize-6,keyUDP,6)!=0) )
		{	
		if (dbg) printf("Invalid KEY :%s\n",udpBuffer);	
		recvMsgSize=1;
		}
	time (&simdi1);
	//NS+ ECMPID + SID =16 +32 +6
	if (recvMsgSize==54)
		{
		pthread_mutex_lock( &dcwmutex );
		curdcw=dcwcnt;
		for (ix=0;ix<dcwcnt;ix++)
			{
			if (memcmp(udpBuffer,dcwlist[ix].chan,16)==0)
				{
				curdcw=ix; 
				break;
				}
			}
		if (curdcw==dcwcnt) dcwcnt+=1;
		if (dcwcnt==MAXDCW) { curdcw=0; dcwcnt=1; }
			
		memcpy(dcwlist[curdcw].chan,udpBuffer,16);
		memcpy(dcwlist[curdcw].keys,udpBuffer+16,32);
		time (&dcwlist[curdcw].zaman);
		dcwlist[curdcw].chan[16]=0;
		dcwlist[curdcw].keys[32]=0;
		if (dbg) printf("DCW Key Received %d  : %s%s from %s\n",curdcw,dcwlist[curdcw].chan,dcwlist[curdcw].keys,inet_ntoa(udpClntAddr.sin_addr));
		pthread_mutex_unlock( &dcwmutex );			//unlock mutex
		} //msgsize-54 if
    	} // for ever
    return (NULL);
}



void *Srv_status(void * )
{
int 	x;
struct 	userlist *userl;
char 	log[200],html[512];
FILE *  logf;
FILE *  logh;
time_t rawtime;

pthread_detach(pthread_self()); 

while (1)
	{
	sleep(10);
	for (x=0;x<Ncnt;x++) 
		{
		pthread_mutex_lock( &nccmutex[x] );
	   	if (!Ncc[x]->is_ready())
			{
			Nlist[x].active=false;
			delete(Ncc[x]);
			printf("------------INIT NewCamd %d ------------------------------\n",x);
			if (Ncc[x] = new newcamd_client(Nlist[x].confn,timeout,VERSION_525), !Ncc[x])
				{
				perror("NEWCAMD Server #%d  Init Failed ");
				return(0);
				}
			Ncc[x]->dbg=dbg;
			Nlist[x].active=Ncc[x]->is_ready();
			}
		pthread_mutex_unlock( &nccmutex[x] );
		}

	for (x=0;x<Ccnt;x++) 
		{
		pthread_mutex_lock( &cccmutex[x] );
		if (!Ccc[x]->ready)
			{
			Clist[x].active=false;
			delete(Ccc[x]);
			printf("------------INIT CCcam %d ------------------------------\n",x);
			if (Ccc[x] = new cCardClientCCcam2(Clist[x].confn,timeout,Clist[x].hop), !Ccc[x])
				{
				perror("CCcam Client init Failed.");
				return(0);
				}
			Clist[x].active=Ccc[x]->ready;
			Ccc[x]->dbg = dbg;
			}
		pthread_mutex_unlock( &cccmutex[x] );
		}

	//Http post data
	strcpy(log,apath);
	strcat(log,"btksrv.log");
	logf=fopen(log,"w");
	//if (logf==NULL) continue;

	strcpy(html,apath);
	strcat(html,"btksrv.txt");
	logh=fopen(html,"w");
	strcpy(html,"------------ Global Statistics ----------\n");
	fputs(html,logh);

	time ( &rawtime );
	//if (dbg) printf("*** LOG is written to %s on %s\n",log,ctime(&rawtime));
	pthread_mutex_lock( &usrmutex );
	userl=users;
	sprintf(log,"Sum=%d,%d,%d,%d,%d",stat_ecm_total,stat_ecm_ok,stat_ecm_nok,stat_ecm_cache,Ccnt+Ncnt);
	fputs(log,logf);
	if (stat_ecm_total==0) stat_ecm_total=1;
	//sprintf(html,"Server Uptime     : %d\n",conn); fputs(html,logh);
	sprintf(html,"Active User Count : %d\n",conn); fputs(html,logh);
	sprintf(html,"Ecm Sent Total    : %d\n",stat_ecm_total); fputs(html,logh);
	sprintf(html,"Ecm Sent OK       : %d \t %%%d\n",stat_ecm_ok,100*stat_ecm_ok/stat_ecm_total); fputs(html,logh);
	sprintf(html,"Ecm Sent NotOk    : %d \t %%%d\n",stat_ecm_nok,100*stat_ecm_nok/stat_ecm_total); fputs(html,logh);
	sprintf(html,"Ecm Sent Cache    : %d \t %%%d\n\n",stat_ecm_cache,100*stat_ecm_cache/stat_ecm_total); fputs(html,logh);
	
	//if (Ccnt>0)  { strcpy(html,"\n----------------------- CCcamd Lines -------------------------\n\n"); fputs(html,logh); }
	for (x=0;x<Ccnt;x++) 
		{
		sprintf(log,",C[%d],%d,%d,%d",x,Clist[x].active,Clist[x].stat.ecm,Clist[x].stat.ecmok);
		fputs(log,logf);
		sprintf(html,"CCcamd  Line [%d] -> Status: %d \t ECM Total: %d \t ECM Ok: %d \t ECM NOK: %d\n\n",
					  x, Clist[x].active, Clist[x].stat.ecm, Clist[x].stat.ecmok, Clist[x].stat.ecm-Clist[x].stat.ecmok);
		fputs(html,logh);
		}

	//if (Ncnt>0) { strcpy(html,"\n------------------------ Newcamd Lines -----------------------\n\n"); fputs(html,logh); }
	for (x=0;x<Ncnt;x++) 
		{
		sprintf(log,",N[%d],%d,%d,%d",x,Nlist[x].active,Nlist[x].stat.ecm,Nlist[x].stat.ecmok);
		fputs(log,logf);
		sprintf(html,"Newcamd Line [%d] -> Status: %d \t ECM Total: %d \t ECM Ok: %d \t ECM NOK: %d\n\n",
					  x, Nlist[x].active, Nlist[x].stat.ecm, Nlist[x].stat.ecmok, Nlist[x].stat.ecm-Nlist[x].stat.ecmok);
		fputs(html,logh);
		}

	strcpy(html,"\n"); fputs(html,logh);
	strcpy(html,"----SN-----USERNAME-----------UP-----VER-----IP_ADDRESS----------Total-----Succ-----Fail-----Cache----\n"); 
	fputs(html,logh);

	int cpos,cloc;
	int loc[10]; //0-sn  1-user,  2-login  3-vers  4-ip  Ecm  5-tot  6-OK   7-Nok  8-cache
	char *pos;
	char tmps[30];
	pos=html;
	cpos=0; cloc=0;
	while (pos[cpos]!='\n')
		{
		if (pos[cpos]!='-')
			{
			//printf("Found: %d\n",cpos);
			loc[cloc++]=cpos;
			while ( (pos[cpos]!='-') && (pos[cpos]!='\n') ) cpos++;
			}
		else cpos++;
		}
	//printf("Done\n");
	cpos=strlen(html);
	memset(html,32,cpos-1);
	pos[cpos-1]='\n';
	pos[cpos]='\0';
	fputs(html,logh);
	//
	x=0;
	while (userl!=NULL)
		{
		x++;
		sprintf(log,"&VD%d=%s,%s,%s,%d,%d,%d,%d,%d",x,userl->user,userl->ip,userl->cver,userl->ok,
			userl->ecmtotal,userl->ecmok,userl->ecmnok,userl->ecmcache);
		fputs(log,logf);
		
		//x, userl->user, userl->ok, userl->ip, userl->cver, userl->ecmtotal, userl->ecmok, userl->ecmnok, userl->ecmcache);
		memset(html,32,cpos-1);
		sprintf(tmps,"%d",x); 			memcpy(pos+loc[0],tmps,strlen(tmps));
		sprintf(tmps,"%s",userl->user); 	memcpy(pos+loc[1],tmps,strlen(tmps));
		sprintf(tmps,"%d",userl->ok); 		memcpy(pos+loc[2],tmps,strlen(tmps));
		sprintf(tmps,"%s",userl->cver); 	memcpy(pos+loc[3],tmps,strlen(tmps));
		sprintf(tmps,"%s",userl->ip); 		memcpy(pos+loc[4],tmps,strlen(tmps));
		sprintf(tmps,"%d",userl->ecmtotal); 	memcpy(pos+loc[5],tmps,strlen(tmps));
		sprintf(tmps,"%d",userl->ecmok); 	memcpy(pos+loc[6],tmps,strlen(tmps));
		sprintf(tmps,"%d",userl->ecmnok); 	memcpy(pos+loc[7],tmps,strlen(tmps));
		sprintf(tmps,"%d",userl->ecmcache); 	memcpy(pos+loc[8],tmps,strlen(tmps));
		fputs(html,logh);
		userl=userl->next;
		}
	fclose(logf);
	fclose(logh);
	pthread_mutex_unlock( &usrmutex );
	}
	
return (NULL);
}

#ifdef _CSP
// CSP activation
void *CSP_runner(void * )
{
pthread_detach(pthread_self()); 
while (mycsp->ready) mycsp->main();
pthread_exit(NULL);
}
#endif

